/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements. See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership. The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.xenei.jdbc4sparql.iface;

import java.util.Collection;
import java.util.Iterator;

/**
 * Filters a namespacedObject by name.
 * 
 * @param <T>
 *            a NamespacedObject
 */
public class NameFilter<T extends NamedObject> implements Iterator<T>,
		Iterable<T>
{
	// the name pattern to match
	private String namePattern;
	// the iterator of the original collection.
	private Iterator<? extends T> iter;
	// our next object.
	private T next;

	/**
	 * Construct a NameFilter from a pattern and a collection.
	 * 
	 * If namePattern is null match all names.
	 * 
	 * @param namePattern
	 *            The pattern to match or null.
	 * @param objs
	 *            The collection of objects to filter.
	 */
	public NameFilter( final String namePattern,
			final Collection<? extends T> objs )
	{
		this(namePattern, objs.iterator());
	}

	/**
	 * Construct a NameFilter from a pattern and an iterator.
	 * 
	 * If namePattern is null match all names.
	 * 
	 * @param namePattern
	 *            The pattern to match or null.
	 * @param iter
	 *            the iterator of objects to filter.
	 */
	public NameFilter( final String namePattern,
			final Iterator<? extends T> iter )
	{
		this.namePattern = namePattern;
		this.iter = iter;
		next = null;
	}

	@Override
	public boolean hasNext()
	{
		if (namePattern == null)
		{
			return iter.hasNext();
		}

		while ((next == null) && iter.hasNext())
		{
			next = iter.next();
			if (!next.getName().equals(namePattern))
			{
				next = null;
			}
		}
		return next != null;
	}

	@Override
	public Iterator<T> iterator()
	{
		return this;
	}

	@Override
	public T next()
	{
		if (namePattern == null)
		{
			return iter.next();
		}
		else
		{
			final T retval = next;
			next = null;
			return retval;
		}
	}

	@Override
	public void remove()
	{
		throw new UnsupportedOperationException();
	}

}
